mod datastore;
mod grpc;
mod http_server;
use crate::datastore::LOCALHOST;
use crate::grpc::challenge::NodeUpdate;
use datastore::Store;
use solana_sdk::signature::keypair::Keypair;
use solana_sdk::signer::Signer;
use std::fs::File;
use std::io::{BufRead, BufReader};
use std::sync::Arc;
use tokio::sync::broadcast;
use tokio::sync::broadcast::Sender;
use tokio::time::{sleep, Duration};

use tokio::task::JoinSet;

const INIT_NODES: &str = "detee_challenge_nodes";

pub async fn localhost_cron(ds: Arc<Store>, tx: Sender<NodeUpdate>) {
    loop {
        sleep(Duration::from_secs(60)).await;
        let _ = tx.send((LOCALHOST.to_string(), ds.get_localhost()).into());
        ds.remove_inactive_nodes();
    }
}

async fn get_keypair() -> Keypair {
    let input = match File::open(INIT_NODES) {
        Ok(i) => i,
        Err(_) => {
            println!("Could not find remote nodes in the file {INIT_NODES}");
            println!("Starting a new network with a new key...");
            return Keypair::new();
        }
    };
    let buffered = BufReader::new(input);
    for line in buffered.lines() {
        match grpc::client::key_grabber(line.unwrap()).await {
            Ok(keypair) => {
                println!(
                    "Got keypair from the network. Joining the network using wallet {}.",
                    keypair.pubkey()
                );
                return keypair;
            }
            Err(e) => {
                println!("Could not get keypair: {e:?}");
            }
        };
    }
    panic!("could not get keypair.");
}

#[tokio::main]
async fn main() {
    let keypair = get_keypair().await;
    let ds = Arc::new(Store::init(keypair));

    let (tx, mut _rx) = broadcast::channel(500);

    let mut long_term_tasks = JoinSet::new();
    let mut init_tasks = JoinSet::new();

    long_term_tasks.spawn(localhost_cron(ds.clone(), tx.clone()));
    long_term_tasks.spawn(http_server::init(ds.clone()));
    long_term_tasks.spawn(grpc::server::MyServer::init(ds.clone(), tx.clone()).start());

    if let Ok(input) = File::open(INIT_NODES) {
        let buffered = BufReader::new(input);
        for line in buffered.lines() {
            init_tasks.spawn(
                grpc::client::ConnManager::init(ds.clone(), tx.clone())
                    .start_with_node(line.unwrap()),
            );
        }
    }

    let mut connection_count = 0;
    while init_tasks.join_next().await.is_some() {
        if connection_count < 3 {
            long_term_tasks.spawn(grpc::client::ConnManager::init(ds.clone(), tx.clone()).start());
            connection_count += 1;
        }
    }

    while connection_count < 3 {
        long_term_tasks.spawn(grpc::client::ConnManager::init(ds.clone(), tx.clone()).start());
        connection_count += 1;
    }

    // exit no matter which task finished
    long_term_tasks.join_next().await;

    println!("Shutting down...");
}
